import 'package:flutter/material.dart';
import 'package:path_to_regexp/path_to_regexp.dart';

class PathStackEntry {
  PathStackEntry({required this.path, required this.builder, this.maintainState = true, this.aliases});
  final String path;
  final Widget Function(BuildContext context, PathArgs params) builder;
  final List<String>? aliases;
  final bool maintainState;
  // Paths ending with a / are assumed to allow prefixes on the end
  bool get allowSuffix => path.endsWith("/");
}

class PathStack extends StatefulWidget {
  static const kDefaultDuration = Duration(milliseconds: 250);

  final String path;
  final String parentPath;
  final List<PathStackEntry> entries;
  final bool caseSensitive;
  final Widget Function(BuildContext context, Widget child)? scaffoldBuilder;
  final Widget Function(BuildContext context)? unknownPathBuilder;
  final Widget Function(BuildContext context, Widget child, AnimationController animation)? transitionBuilder;
  final Duration transitionDuration;

  PathStack(
      {Key? key,
      required this.path,
      required this.entries,
      this.parentPath = "",
      this.caseSensitive = false,
      this.scaffoldBuilder,
      this.transitionBuilder,
      this.transitionDuration = kDefaultDuration,
      this.unknownPathBuilder})
      : super(key: key);

  static PathStackState of(BuildContext context) =>
      (context.dependOnInheritedWidgetOfExactType<_InheritedPathStack>() as _InheritedPathStack).state;

  @override
  PathStackState createState() => PathStackState();
}

class PathStackState extends State<PathStack> with SingleTickerProviderStateMixin {
  Map<String, Widget> _knownRoutes = {};
  String? _previousMatch;
  late AnimationController transitionInAnim = AnimationController(vsync: this, duration: widget.transitionDuration)
    ..forward();

  @override
  Widget build(BuildContext context) {
    // Try and find a known route for the current path
    PathStackEntry? matchingRoute =
        List<PathStackEntry?>.from(widget.entries).firstWhere(checkEntryMatchesPath, orElse: () => null);
    // Fall back to first page in the route stack, TODO: Add a emptyRouteBuilder?
    if (matchingRoute == null) {
      print("WARNING: Unable to find a route for ${widget.path}");
      matchingRoute = PathStackEntry(
          path: "404",
          builder: (c, _) {
            return widget.unknownPathBuilder?.call(c) ?? Center(child: Text("Page Not Found"));
          });
    }
    // Use a UniqueKey if maintainState=false
    Key pageKey = matchingRoute.maintainState ? ValueKey(matchingRoute.path) : UniqueKey();
    // Add the new route to our list of known routes (if maintainState = false, we'll remove it later)
    _knownRoutes[matchingRoute.path] = _KeyedWidget(
      key: pageKey,
      child: matchingRoute.builder.call(
        context,
        PathArgs.fromPath(route: matchingRoute.path, path: widget.path),
      ),
    );
    // Get all known children which we'll pass to the indexedStack
    List<Widget> children = _knownRoutes.values.toList();
    // Finds the index of the matching route, which we'll also pass to indexedStack
    int index = _knownRoutes.keys.toList().indexWhere((r) => r == matchingRoute!.path);
    // Now that we have the index and list of children, remove this route from the cache if maintainState = false
    if (matchingRoute.maintainState == false) {
      _knownRoutes.remove(matchingRoute.path);
    }
    // Trigger transitionIn anim
    if (_previousMatch != matchingRoute.path) {
      transitionInAnim.forward(from: 0);
    }
    _previousMatch = matchingRoute.path;

    // Wrap builders
    Widget content = IndexedStack(index: index, children: children);
    content = widget.transitionBuilder?.call(context, content, transitionInAnim) ?? content;
    content = widget.scaffoldBuilder?.call(context, content) ?? content;
    return _InheritedPathStack(child: content, state: this);
  }

  void clearState() => _knownRoutes.clear();

  @override
  void dispose() {
    clearState();
    super.dispose();
  }

  bool checkEntryMatchesPath(PathStackEntry? entry) {
    List<String> allPaths = List.from([entry!.path, ...(entry.aliases ?? [])]);
    for (var i = 0; i < allPaths.length; i++) {
      final regExp = pathToRegExp('${widget.parentPath}${allPaths[i]}', prefix: entry.allowSuffix);
      if (regExp.hasMatch("${widget.path}")) return true;
    }
    return false;
  }
}

class PathArgs {
  PathArgs({required this.data});
  final Map<String, String> data;

  static PathArgs fromPath({required String route, required String path}) {
    // Get path params
    List<String> paramList = [];
    final regExp = pathToRegExp(route, parameters: paramList);
    Map<String, String> pathMap = {};
    if (regExp.hasMatch(path)) {
      final match = regExp.matchAsPrefix(path)!;
      pathMap = extract(paramList, match);
    }
    // Get queryString params
    Map<String, String> queryMap = Uri.parse(path).queryParameters;

    return PathArgs(data: {...queryMap, ...pathMap});
  }
}

class _KeyedWidget extends StatelessWidget {
  const _KeyedWidget({Key? key, required this.child}) : super(key: key);
  final Widget child;
  @override
  Widget build(BuildContext context) => child;
}

class _InheritedPathStack extends InheritedWidget {
  _InheritedPathStack({Key? key, required Widget child, required this.state}) : super(key: key, child: child);
  final PathStackState state;
  @override
  bool updateShouldNotify(covariant InheritedWidget oldWidget) => true;
}
